"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const type_analyzer_1 = require("@previewjs/type-analyzer");
const vfs_1 = require("@previewjs/vfs");
const path_1 = __importDefault(require("path"));
const vitest_1 = require("vitest");
const _1 = require(".");
const analyze_component_1 = require("./analyze-component");
const vue_reader_1 = require("./vue-reader");
const ROOT_DIR_PATH = path_1.default.join(__dirname, "virtual");
const MAIN_FILE = path_1.default.join(ROOT_DIR_PATH, "App.vue");
const EMPTY_SET = new Set();
(0, vitest_1.describe)("analyze Vue 3 component", () => {
    let memoryReader;
    let typeAnalyzer;
    let frameworkPlugin;
    (0, vitest_1.beforeEach)(async () => {
        memoryReader = (0, vfs_1.createMemoryReader)();
        frameworkPlugin = await _1.vue3FrameworkPlugin.create();
        typeAnalyzer = (0, type_analyzer_1.createTypeAnalyzer)({
            rootDirPath: ROOT_DIR_PATH,
            reader: (0, vue_reader_1.createVueTypeScriptReader)((0, vfs_1.createStackedReader)([
                memoryReader,
                (0, vfs_1.createFileSystemReader)({
                    watch: false,
                }),
                (0, vfs_1.createFileSystemReader)({
                    mapping: {
                        from: path_1.default.join(__dirname, "..", "preview", "modules"),
                        to: path_1.default.join(ROOT_DIR_PATH, "node_modules"),
                    },
                    watch: false,
                }),
            ])),
            tsCompilerOptions: frameworkPlugin.tsCompilerOptions,
        });
    });
    (0, vitest_1.afterEach)(() => {
        typeAnalyzer.dispose();
    });
    (0, vitest_1.test)("defineProps<Props>()", async () => {
        (0, vitest_1.expect)(await analyze(`
<script setup lang="ts">
defineProps<{ foo: string }>();
</script>

<template>
  <div>
    Hello, World!
  </div>
</template>
`)).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("withDefaults(defineProps<Props>(), ...)", async () => {
        (0, vitest_1.expect)(await analyze(`
<script setup lang="ts">
withDefaults(defineProps<{ foo: string, bar: string }>(), {
  foo: "hello"
});
</script>

<template>
  <div>
    Hello, World!
  </div>
</template>
`)).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: (0, type_analyzer_1.optionalType)(type_analyzer_1.STRING_TYPE),
                bar: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("const props = defineProps<Props>()", async () => {
        (0, vitest_1.expect)(await analyze(`
<script setup lang="ts">
const props = defineProps<{ foo: string }>();
</script>

<template>
  <div>
    Hello, World!
  </div>
</template>
`)).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("const props = withDefaults(defineProps<Props>(), ...)", async () => {
        (0, vitest_1.expect)(await analyze(`
<script setup lang="ts">
const props = withDefaults(defineProps<{ foo: string, bar: string }>(), {
  foo: "hello"
});
</script>

<template>
  <div>
    Hello, World!
  </div>
</template>
`)).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: (0, type_analyzer_1.optionalType)(type_analyzer_1.STRING_TYPE),
                bar: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("props = defineProps<Props>()", async () => {
        (0, vitest_1.expect)(await analyze(`
<script setup lang="ts">
let props;
props = defineProps<{ foo: string }>();
</script>

<template>
  <div>
    Hello, World!
  </div>
</template>
`)).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("props = withDefaults(defineProps<Props>(), ...)", async () => {
        (0, vitest_1.expect)(await analyze(`
<script setup lang="ts">
let props;
props = withDefaults(defineProps<{ foo: string, bar: string }>(), {
  foo: "hello"
});
</script>

<template>
  <div>
    Hello, World!
  </div>
</template>
`)).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: (0, type_analyzer_1.optionalType)(type_analyzer_1.STRING_TYPE),
                bar: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("export default defineComponent() simple case", async () => {
        (0, vitest_1.expect)(await analyze(`
<template>
  <div>{{ label }}</div>
</template>
<script lang="ts">
import { defineComponent } from "vue";

export default defineComponent({
  props: {
    foo: String,
    bar: { type: String, required: true },
  },
});
</script>
`)).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: (0, type_analyzer_1.optionalType)(type_analyzer_1.STRING_TYPE),
                bar: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: vitest_1.expect.anything(),
        });
    });
    (0, vitest_1.test)("export default defineComponent() complex case", async () => {
        // Source: https://github.com/storybookjs/storybook/blob/4aa1f9944c9ede050c23afcc6861acf99cf5e841/examples/vue-3-cli/src/stories/Button.vue
        (0, vitest_1.expect)(await analyze(`
  <template>
    <div>{{ label }}</div>
  </template>

  <script lang="typescript">
  import { reactive, computed, defineComponent } from 'vue';

  export default defineComponent({
    name: 'App',

    props: {
      label: {
        type: String,
        required: true,
      },
      sublabel: {
        type: String,
        default: 'sublabel',
      },
      primary: {
        type: Boolean,
        default: false,
      },
      size: {
        type: String,
        validator: function (value) {
          return ['small', 'medium', 'large'].indexOf(value) !== -1;
        },
      },
      backgroundColor: {
        type: String,
      },
    },

    emits: ['click'],

    setup(props, { emit }) {
      props = reactive(props);
      return {
        style: computed(() => ({
          backgroundColor: props.backgroundColor,
        })),
        onClick() {
          emit('click');
        }
      }
    },
  });
  </script>
  `)).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                label: type_analyzer_1.STRING_TYPE,
                sublabel: (0, type_analyzer_1.optionalType)(type_analyzer_1.STRING_TYPE),
                primary: (0, type_analyzer_1.optionalType)(type_analyzer_1.BOOLEAN_TYPE),
                size: (0, type_analyzer_1.optionalType)(type_analyzer_1.STRING_TYPE),
                backgroundColor: (0, type_analyzer_1.optionalType)(type_analyzer_1.STRING_TYPE),
            }),
            providedArgs: EMPTY_SET,
            types: vitest_1.expect.anything(),
        });
    });
    async function analyze(source) {
        memoryReader.updateFile(MAIN_FILE, source);
        return (0, analyze_component_1.analyzeVueComponentFromTemplate)(typeAnalyzer, MAIN_FILE);
    }
});
