"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createVueTypeScriptReader = void 0;
const vfs_1 = require("@previewjs/vfs");
const compiler_sfc_1 = require("@vue/compiler-sfc");
const path_1 = __importDefault(require("path"));
const typescript_1 = __importDefault(require("typescript"));
/**
 * Returns a reader that returns a virtual TypeScript file
 * named foo.vue.ts for every corresponding foo.vue file.
 */
function createVueTypeScriptReader(reader) {
    return new VueTypeScriptReader(reader);
}
exports.createVueTypeScriptReader = createVueTypeScriptReader;
class VueTypeScriptReader {
    constructor(reader) {
        var _a;
        this.reader = reader;
        this.listeners = new vfs_1.ReaderListeners();
        this.observe = (_a = this.reader.observe) === null || _a === void 0 ? void 0 : _a.bind(this.reader);
        reader.listeners.add({
            onChange: (absoluteFilePath, info) => {
                this.listeners.notify(absoluteFilePath, info);
            },
        });
    }
    async read(absoluteFilePath) {
        if (absoluteFilePath.endsWith(".vue.ts")) {
            const source = await this.reader.read(absoluteFilePath.substr(0, absoluteFilePath.length - 3));
            if ((source === null || source === void 0 ? void 0 : source.kind) !== "file") {
                return null;
            }
            const name = path_1.default.basename(absoluteFilePath);
            return {
                kind: "file",
                name,
                realPath: async () => null,
                lastModifiedMillis: () => source.lastModifiedMillis(),
                read: async () => convertToTypeScript(await source.read(), name),
                size: () => source.size(),
            };
        }
        return this.reader.read(absoluteFilePath);
    }
    readSync(absoluteFilePath) {
        if (absoluteFilePath.endsWith(".vue.ts")) {
            const source = this.reader.readSync(absoluteFilePath.substr(0, absoluteFilePath.length - 3));
            if ((source === null || source === void 0 ? void 0 : source.kind) !== "file") {
                return null;
            }
            const name = path_1.default.basename(absoluteFilePath);
            return {
                kind: "file",
                name,
                realPath: () => null,
                lastModifiedMillis: () => source.lastModifiedMillis(),
                read: () => convertToTypeScript(source.read(), name),
                size: () => source.size(),
            };
        }
        return this.reader.readSync(absoluteFilePath);
    }
}
function convertToTypeScript(vueTemplateSource, name) {
    const parsed = (0, compiler_sfc_1.parse)(vueTemplateSource);
    if (parsed.descriptor.scriptSetup) {
        return `import { defineProps } from '@vue/runtime-core';\n${parsed.descriptor.scriptSetup.content}`;
    }
    else if (parsed.descriptor.script) {
        const lang = parsed.descriptor.script.lang;
        if (lang && !["js", "javascript", "ts", "typescript"].includes(lang)) {
            return "";
        }
        const scriptContent = parsed.descriptor.script.content;
        const sourceFile = typescript_1.default.createSourceFile(`${name}.vue.ts`, scriptContent, typescript_1.default.ScriptTarget.Latest);
        const transformation = typescript_1.default.transform(sourceFile, [
            (context) => (sourceFile) => {
                if (typescript_1.default.isSourceFile(sourceFile)) {
                    return typescript_1.default.visitEachChild(sourceFile, (node) => {
                        if (typescript_1.default.isExportAssignment(node)) {
                            return typescript_1.default.factory.createVariableStatement([], typescript_1.default.factory.createVariableDeclarationList([
                                typescript_1.default.factory.createVariableDeclaration("pjs_component", undefined, undefined, typescript_1.default.factory.createAsExpression(extractDefineComponentArgument(node.expression), typescript_1.default.factory.createTypeReferenceNode("const"))),
                            ], typescript_1.default.NodeFlags.Const));
                        }
                        return node;
                    }, context);
                }
                return sourceFile;
            },
        ]);
        const transformed = typescript_1.default
            .createPrinter()
            .printFile(transformation.transformed[0]);
        transformation.dispose();
        return `
${transformed}

import type { PropType as PJS_PropType } from "@vue/runtime-core";

type PJS_TypeOrUnion<T> = PJS_PropType<T> | ReadonlyArray<PJS_PropType<T>>;
type PJS_OptionalPropType<T> = PJS_TypeOrUnion<T> | {type: PJS_TypeOrUnion<T>; required?: false};
type PJS_RequiredPropType<T> = {type: PJS_TypeOrUnion<T>; required: true};
type PJS_OptionalPropsKeys<T> = {
  [K in keyof T]: T[K] extends PJS_OptionalPropType<any> ? K : never;
}[keyof T];
type PJS_RequiredPropsKeys<T> = {
  [K in keyof T]: T[K] extends PJS_RequiredPropType<any> ? K : never;
}[keyof T];
type PJS_CombinedProps<T> = T extends readonly [...any] ? {
  [K in T[number]]: unknown
} : ({
  [K in PJS_OptionalPropsKeys<T>]?: T[K] extends PJS_OptionalPropType<infer S> ? S : never;
} & {
  [K in PJS_RequiredPropsKeys<T>]: T[K] extends PJS_RequiredPropType<infer S> ? S : never;
});
type PJS_ExtractProps<T> = T extends { props: any } ? PJS_CombinedProps<T['props']> : {}
type PJS_Props = PJS_ExtractProps<typeof pjs_component>;
`;
    }
    else {
        return "";
    }
}
function extractDefineComponentArgument(node) {
    if (typescript_1.default.isCallExpression(node) &&
        typescript_1.default.isIdentifier(node.expression) &&
        node.expression.text === "defineComponent" &&
        node.arguments.length > 0) {
        return node.arguments[0];
    }
    return node;
}
